#ifndef lint
static char sccsid[] = "@(#)execute.c	1.10 (UKC) 9/7/92";
#endif /* !lint */

/* 
 * Copyright 1991,1992 Richard Hesketh / rlh2@ukc.ac.uk
 *                     Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Richard Hesketh AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Richard Hesketh OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */

/* Contains all the routines for forking and executing Ghostscript interpreter
 * Three pipes are opened up between "gs" and "gspreview" and
 * the standard output and standard errors generated by the script are sent down
 * these pipes.  The third pipe is used to send messages to "gs" such as ..
 * "(/tmp/gsp24354.0)run" to interpret a Ghostscript file.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <ctype.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xauth.h>
#include "resources.h"

#define GS_MSG_MARKER '\001'
#define GS_ERROR_MARKER '\002'

/* define some extern tools and where they can be found .. override these
 * in the Makefile.
 */
#ifndef GHOSTSCRIPT_INTERPRETER
#define GHOSTSCRIPT_INTERPRETER "/usr/local/lib/X11/gspreview/gs"
#endif

#ifndef GS_HOME
#define GS_HOME "/usr/local/lib/X11/gspreview"
#endif

#ifndef GS_LIB
#define GS_LIB "/usr/local/lib/X11/gspreview:/usr/local/lib/X11/gspreview/fonts"
#endif

#ifndef ZCAT
#define ZCAT "/usr/ucb/zcat"
#endif

#ifndef SH
#define SH "/bin/sh"
#endif

#ifndef DIED_MESG
#define DIED_MESG "\n\n**** Ghostscript interpreter died unexpectedly ****\n"
#endif

#ifndef ADJUST_MESG
#define ADJUST_MESG "\n\n**** X Server is running low on free memory and\n**** denied requested page size,\n**** so using largest allowed size.\n\n"
#endif

static Boolean	xio_error_happened = FALSE;
static Boolean	restart_upon_exit = FALSE;
static Boolean	has_finished = FALSE;
static Boolean	both_pipes_closed = FALSE;
static XtInputId	error_pipe_id = NULL;
static XtInputId	output_pipe_id = NULL;
static XtInputId	stdin_id = NULL;
static int	gs_pid = NULL;
static int	error_pipe = NULL;
static int	output_pipe = NULL;
static int	input_pipe = NULL;
static int	errpfd[2];
static int	outpfd[2];
static int	inpfd[2];
static Pixmap background_pixmap = None;

extern void OutputText();
extern void ParseError();
extern void OutputError();
extern void BusyProcessing();
extern void QuitGSP();
extern void RereadFile();
extern void LoadFile();
extern void InitializePageFiles();
extern void RemovePageProcessingTimeout();
extern Boolean GotoDirectory();

extern char	*getenv(/* char* */);
extern char	*malloc(/* int */);
extern char	*index();
extern int	getpid();

extern Widget WI_ps_window;
extern String zcat_file;

void CloseStdin();


/*  return sys_errlist[errno] if in range
 */
char *
reason()
{
	extern char *sys_errlist[];
	extern int errno, sys_nerr;

	return((errno > 0 && errno < sys_nerr) ? sys_errlist[errno]
							: "unknown reason");
}


/*
 *	Pick up children
 */
#ifdef SVR4
/* System V R4 additions by David Wexelblat (goblin!dwex@att.att.com) */
static void
child_done()
{
	int status;
	pid_t pid;

	while ((pid = waitpid((pid_t)-1, &status, WNOHANG)) > 0) {
		if (pid == gs_pid) {
			gs_pid = NULL;
		}
	}
}
#else
static void
child_done()
{
	union wait status;
	register int pid;
	
	while ((pid = wait3(&status, WNOHANG, (struct rusage *)NULL)) > 0)
		if (pid == gs_pid) {
			gs_pid = NULL;
		} 
}
#endif


static void
initialize_signal_handlers()
{
	/* clean up when GS dies or is knocked on the head */
	(void)signal(SIGCHLD, child_done);
	/* clean up when we die */
	(void)signal(SIGINT, QuitGSP);
	(void)signal(SIGHUP, RereadFile);
	(void)signal(SIGTERM, QuitGSP);
	(void)signal(SIGQUIT, QuitGSP);
	(void)signal(SIGILL, QuitGSP);
	(void)signal(SIGTRAP, QuitGSP);
	(void)signal(SIGABRT, QuitGSP);
	(void)signal(SIGEMT, QuitGSP);
	(void)signal(SIGFPE, QuitGSP);
	(void)signal(SIGBUS, QuitGSP);
	(void)signal(SIGSEGV, QuitGSP);
	(void)signal(SIGSYS, QuitGSP);
	/* ignore the broken pipe */
	(void)signal(SIGPIPE, SIG_IGN);
}


static void
my_system(command)
String command;
{
	int pid;
	int result;

	initialize_signal_handlers();

	(void)signal(SIGCHLD, SIG_DFL);

	switch (pid = fork()) {
		case -1:
			fprintf(stderr, "%s: Cannot fork\n", PROG_NAME);
			exit(1);
		case 0:
			/*
			 *	Child
			 */
#if defined(SVR4) || defined(SYSV)
			setpgrp();
#else
			setpgrp(0, getpid());
#endif
			
			execlp(SH, "sh", "-c", command, (char *)0);
			exit(1);
		default:
			/* wait here until child dies/completes */
			while(wait(&result) != pid);
			break;
	}
}


static void
finished_executing()
{
	close(output_pipe);
	output_pipe = NULL;
	close(error_pipe);
	error_pipe = NULL;
	close(input_pipe);
	input_pipe = NULL;

	BusyProcessing(restart_upon_exit);

	/* Ghostscript has finished and all the pipes have been closed */
	if (!restart_upon_exit) {
		/* Ghostscript died for some reason so say so ... */
		char *tmp = XtNewString(DIED_MESG);
		ParseError(tmp);
		XtFree(tmp);
		RemovePageProcessingTimeout();
	}
}


/* output from the script is available on one of two pipes .. the stdout
 * pipe and the stderr pipe.   Drain the pipe a bit at a time unless the
 * script has finished in which case drain the whole of the pipe.
 * When both pipes have been drained they are closed and the script is
 * deemed to have stop executing and finished_executing() is called.
 */
static void
process_output(client_data, source, id)
XtPointer client_data;
int *source;
XtInputId *id;
{
#define READ_SIZE 512
	char results[READ_SIZE+1];
	int num_read, i;

	do {
		results[0] = '\0';
		num_read = read(*source, results, READ_SIZE);
		if (num_read > 0) {
			results[num_read] = '\0';
		/* convert all unprintable characters in to "@" signs */
		for (i = 0; i < num_read; i++)
			if (!isprint(results[i]) &&
			    results[i] != '\n' &&
			    results[i] != GS_MSG_MARKER &&
			    results[i] != GS_ERROR_MARKER)
				results[i] = '@';

			/* error and stacks shown in error popup */
			ParseError(results);
		} else {
			has_finished = TRUE;
			XtRemoveInput(*id);
			if (id == &output_pipe_id)
				output_pipe_id = NULL;
			else
				error_pipe_id = NULL;
		}
	} while (num_read > 0 && has_finished);

	if (has_finished) {
		if (both_pipes_closed) {
			finished_executing();
			both_pipes_closed = FALSE;
		} else
			both_pipes_closed = TRUE;
	}
}


/* when reading commands from stdin this routine is called to process any
 * waiting input by the toolkit MainLoop.  Basically any text found is
 * simply passed on to the Ghostscript interpreter down its input pipe.
 */
static void
process_stdin(client_data, source, id)
XtPointer client_data;
int *source;
XtInputId *id;
{
#define READ_SIZE 512
	char results[READ_SIZE+1];
	int num_read;
	Boolean has_finished = FALSE;

	do {
		results[0] = '\0';
		num_read = read(*source, results, READ_SIZE);
		if (num_read > 0) {
			results[num_read] = '\0';
			SendCommandToGhostscript(results);
		} else {
			has_finished = TRUE;
			prog_resources.use_stdin = FALSE;
			CloseStdin();
		}
	} while (num_read > 0 && has_finished);
}


static int
xio_handler()
{
	xio_error_happened = TRUE;
	return 1;
}


static void
setup_ghostscript_window(w)
Widget w;
{
	static Atom ghostview_atom = NULL;
	GC clear_gc;
	char buf[256];
	int	x_resolution, y_resolution, width, height, depth;
	int	page_width, page_height;
	int (*old_xio_handler)();

	x_resolution = (int)((double)25.4 * WidthOfScreen(XtScreen(w))
			/ WidthMMOfScreen(XtScreen(w)));
	y_resolution = (int)((double)25.4 * HeightOfScreen(XtScreen(w))
			/ HeightMMOfScreen(XtScreen(w)));
	width = (WidthOfScreen(XtScreen(w)) * prog_resources.page_width)
			/ WidthMMOfScreen(XtScreen(w));
	height = (HeightOfScreen(XtScreen(w)) * prog_resources.page_height)
			/ HeightMMOfScreen(XtScreen(w));

#ifdef DEBUG
printf("Original: width = %d, height = %d\n", width, height);
#endif

	if (prog_resources.page_layout == PAGE_LANDSCAPE ||
	    prog_resources.page_layout == PAGE_SEASCAPE) {
		int swap_dim = width;
		width = height;
		height = swap_dim;
	}

	if (prog_resources.resolution < 0) {
		double percentage;

		percentage = (double)-prog_resources.resolution / 100;
		x_resolution *= percentage;
		y_resolution *= percentage;
		width *= percentage;
		height *= percentage;
	} else if (prog_resources.resolution > 0) {
		double percentage;
		
		percentage = (double)prog_resources.resolution / x_resolution;
		x_resolution = y_resolution = prog_resources.resolution;
		width *= percentage;
		height *= percentage;
	}

	XtUnmanageChild(w);

	if (background_pixmap != NULL) {
		XSetWindowBackgroundPixmap(XtDisplay(w), XtWindow(w), None);
		XFreePixmap(XtDisplay(w), background_pixmap);
		XSync(XtDisplay(w), FALSE);
	}

	XtVaGetValues(w, XtNdepth, &depth, NULL);

	/* have a wrapper around our Pixmap creation so that we can
	 * trap out of memory errors and attempt to use a smaller one.
	 */
	old_xio_handler = XSetErrorHandler(xio_handler);
	xio_error_happened = FALSE;
	background_pixmap = XCreatePixmap(XtDisplay(w), XtWindow(w),
						width, height, depth);
	XSync(XtDisplay(w), FALSE);
	if (xio_error_happened || background_pixmap == NULL) {
		char *adjust_mesg;
		int start_width, start_height;
		int last_w = width, last_h = height;
		int low_h = 0, low_w = 0, high_h = height, high_w = width;

#ifdef DEBUG
printf("Start: width = %d, height = %d\n", width, height);
#endif
		start_width = width;
		start_height = height;

		/* perform a binary chop on the pixmap size to find
		 * one that is closest to the required size without
		 * running out of memory.
		 */
		do {
			if (background_pixmap != NULL) {
				XFreePixmap(XtDisplay(w), background_pixmap);
				XSync(XtDisplay(w), FALSE);
			}

			xio_error_happened = FALSE;

			width = (high_w - low_w) / 2 + low_w;
			height = (high_h - low_h) / 2 + low_h;
#ifdef DEBUG
printf("Trying: width = %d, height = %d\n", width, height);
#endif
			background_pixmap = XCreatePixmap(XtDisplay(w),
								XtWindow(w),
								width, height,
								depth);
			XSync(XtDisplay(w), FALSE);

			if (!xio_error_happened &&
			    background_pixmap != NULL &&
			    (abs(width - last_w) <= 2 &&
			     abs(height - last_h) <= 2))
				/* got a suitable compromise! */
				break;

			if (xio_error_happened) {
#ifdef DEBUG
printf("Too big: width = %d, height = %d\n", width, height);
#endif
				high_w = width;
				high_h = height;
			} else {
#ifdef DEBUG
printf("Too small: width = %d, height = %d\n", width, height);
#endif
				low_w = width;
				low_h = height;
			}
			last_w = width;
			last_h = height;
		} while (1);
#ifdef DEBUG
printf("Finished: width = %d, height = %d\n", width, height);
#endif
		/* now adjust resolution for Ghostscript and inform
		 * user some adjustment had to be made.
		 */
		x_resolution = (int)((double)x_resolution * width / start_width);
		y_resolution = (int)((double)y_resolution * height / start_height);

		adjust_mesg = XtNewString(ADJUST_MESG);
		ParseError(adjust_mesg);
		XtFree(adjust_mesg);
	}
	(void)XSetErrorHandler(old_xio_handler);

	prog_resources.background_pixmap = background_pixmap;
	XSetWindowBackgroundPixmap(XtDisplay(w), XtWindow(w),
				   prog_resources.background_pixmap);
	clear_gc = XCreateGC(XtDisplay(w), XtWindow(w), 0, (XGCValues *)0);
	XSetForeground(XtDisplay(w), clear_gc, WhitePixelOfScreen(XtScreen(w)));
	XFillRectangle(XtDisplay(w), background_pixmap, clear_gc, 0, 0,
			width, height);
	XFreeGC(XtDisplay(w), clear_gc);
	XtVaSetValues(w, XtNwidth, width, XtNheight, height, NULL);
	XtManageChild(w);

	(void)sprintf(buf, "%d %d %d %d %d %d %f %f 0 0 0 0",
			prog_resources.background_pixmap,
			prog_resources.page_layout,
			0, 0,
			(int)((double)prog_resources.page_width * 72.0 / 25.4),
			(int)((double)prog_resources.page_height * 72.0 / 25.4),
			(double)x_resolution, (double)y_resolution);

	if (ghostview_atom == NULL)
		ghostview_atom = XInternAtom(XtDisplay(w), "GHOSTVIEW", 0);

	XChangeProperty(XtDisplay(w), XtWindow(w), ghostview_atom,
			XA_STRING, 8, PropModeReplace,
			(unsigned char*)buf, strlen(buf));
	XSync(XtDisplay(w), FALSE);
}


/************* Module Exported Routines **************/

void
KillGhostscript(restart)
Boolean restart;
{
	restart_upon_exit = restart;

	if (gs_pid != NULL) {
		kill(gs_pid, SIGHUP);
		gs_pid = NULL;

		if (output_pipe_id) {
			XtRemoveInput(output_pipe_id);
			output_pipe_id = NULL;
		}
		if (error_pipe_id) {
			XtRemoveInput(error_pipe_id);
			error_pipe_id = NULL;
		}
	}
	if (restart)
		StartGhostscript();
}


int
StartGhostscript()
{
	int	pid;
	int	stat;
	int	fd;
	fd_set	readfds, fds;
	char	**envp;
	char	*xauth;
	int	i;
	Display *dpy;
	Window win;

	if (gs_pid != NULL)
		return;

	restart_upon_exit = FALSE;
	gs_pid = NULL;
	has_finished = FALSE;
	dpy = XtDisplay(WI_ps_window);
	win = XtWindow(WI_ps_window);

	FD_ZERO(&fds);
	FD_SET(ConnectionNumber(dpy), &fds);

	output_pipe_id = NULL;
	error_pipe_id = NULL;

	both_pipes_closed = FALSE;

	if (pipe(errpfd) == -1) {
		fprintf(stderr, "%s: Cannot open a stderr pipe (%s)\n",
				PROG_NAME, reason());
		return (0);
	}

	if (pipe(outpfd) == -1) {
		fprintf(stderr, "%s: Cannot open a stdout pipe (%s)\n",
				PROG_NAME, reason());
		close(errpfd[0]);
		close(errpfd[1]);
		return (0);
	}

	if (pipe(inpfd) == -1) {
		fprintf(stderr, "%s: Cannot open a stdin pipe (%s)\n",
				PROG_NAME, reason());
		close(errpfd[0]);
		close(errpfd[1]);
		close(outpfd[0]);
		close(outpfd[1]);
		return (0);
	}

	if (output_pipe != NULL)
		close(output_pipe);
	if (error_pipe != NULL)
		close(error_pipe);
	if (input_pipe)
		close(input_pipe);
	output_pipe = outpfd[0];
	error_pipe = errpfd[0];
	input_pipe = inpfd[1];

	output_pipe_id = XtAppAddInput(XtDisplayToApplicationContext(dpy),
					output_pipe, XtInputReadMask,
					process_output, (XtPointer)NULL);
	error_pipe_id = XtAppAddInput(XtDisplayToApplicationContext(dpy),
					error_pipe, XtInputReadMask,
					process_output, (XtPointer)NULL);

	setup_ghostscript_window(WI_ps_window);

	initialize_signal_handlers();

	switch (pid = fork()) {
		case -1:
			fprintf(stderr, "%s: Cannot fork\n", PROG_NAME);
			close(errpfd[0]);
			close(errpfd[1]);
			close(outpfd[0]);
			close(outpfd[1]);
			close(inpfd[0]);
			close(inpfd[1]);
			return 0;
		case 0:
			/*
			 *	Child
			 */
			if (chdir(GS_HOME) < 0)
				(void)GotoDirectory((String)NULL);
#if defined(SVR4) || defined(SYSV)
			setpgrp();
#else
			setpgrp(0, getpid());
#endif
			
			close(ConnectionNumber(dpy));
			close(outpfd[0]);
			close(errpfd[0]);
			close(inpfd[1]);
			if (dup2(inpfd[0], 0) < 0)
				abort();
			if (dup2(outpfd[1], 1) < 0)
				abort();
			if (dup2(errpfd[1], 2) < 0)
				abort();

			envp = (char **)malloc(sizeof(char *) * 5);
			envp[0] = malloc(strlen(DisplayString(dpy)) + 9);
			(void)sprintf(envp[0], "DISPLAY=%s",
						DisplayString(dpy));
			envp[1] = malloc(50);
			(void)sprintf(envp[1], "GHOSTVIEW=%d 0", win);
			envp[2] = malloc(strlen(GS_LIB) + 8);
			(void)sprintf(envp[2], "GS_LIB=%s", GS_LIB);
			xauth = XauFileName();
			if (xauth == NULL)
				envp[3] = (char *)0;
			else {
				envp[3] = malloc(strlen(xauth) + 12);
				(void)sprintf(envp[3], "XAUTHORITY=%s", xauth);
				envp[4] = (char *)0;
			}

			execle(GHOSTSCRIPT_INTERPRETER, "gs", (char *)0, envp);
			close(outpfd[1]);
			close(errpfd[1]);
			close(inpfd[0]);
			exit(1);
		default:
			gs_pid = pid;
			break;
	}
	close(outpfd[1]);
	close(errpfd[1]);
	close(inpfd[0]);

	return (gs_pid);
}


SendCommandToGhostscript(command)
char *command;
{
	if (input_pipe == NULL || command == NULL || !*command)
		return;

	write(input_pipe, command, strlen(command));
}


/* uncompress the given file into a temporary file and then pass
 * back the filename of this temporary uncompressed file.
 */
String
UncompressFile(file)
String file;
{
	String tmp;

	if (zcat_file != NULL)
		(void)unlink(zcat_file);
	else
		InitializePageFiles();

	tmp = XtMalloc(strlen(ZCAT) + strlen(file) + strlen(zcat_file) + 4 + 1);
	(void)sprintf(tmp, "%s %s > %s", ZCAT, file, zcat_file);
	my_system(tmp);
	return (zcat_file);
}


void
OpenStdin(filedes)
int filedes;
{
	if (isatty(filedes)) {
		if (prog_resources.use_stdin) {
			/* read commands interactively from stdin */
			StartGhostscript();
			stdin_id = XtAppAddInput(
				XtWidgetToApplicationContext( WI_ps_window),
				fileno(stdin), XtInputReadMask,
				process_stdin, (XtPointer)NULL);
#ifdef QMS_CLIPPING
			if (prog_resources.picasso)
				SendCommandToGhostscript(QMS_CLIP_PATH);
#endif /* QMS_CLIPPING */
		}
	} else {
		/* we are on the end of a pipe so read everything into
		 * a temporary file and then process it.
		 */
		char buffer[BUFSIZ];
		int num_read, outdes;

		InitializePageFiles();

		if ((outdes = open(zcat_file, O_WRONLY | O_CREAT, 00600)) < 0) {
			fprintf(stderr,
				"%s: Cannot open temporary file %s (%s)\n",
				PROG_NAME, zcat_file, reason());
			exit(1);
		}

		stdin_id = NULL;
		while ((num_read = read(filedes, buffer, BUFSIZ)) > 0)
			write(outdes, buffer, num_read);
		close(outdes);

		prog_resources.use_stdin = FALSE;
		BusyProcessing(TRUE);
		LoadFile(zcat_file);
	}
}


void
CloseStdin()
{
	if (stdin_id == NULL)
		return;
	XtRemoveInput(stdin_id);
	stdin_id = NULL;
	prog_resources.use_stdin = FALSE;
	ResetGhostscript();
}
